import argparse
import logging
import numpy as np
import os
import random
import torch

from tqdm import tqdm


def set_seed(args):
    """Set random seeds for training

    :param args: (argparse.Arguments) arguments passed from the command line.
    """
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    if args.n_gpu > 0:
        torch.cuda.manual_seed_all(args.seed)


def parse_input_arguments(additional_arg_parser):
    """Parses arguments passed from the command line.

    :param additional_arg_parser: (def) A custom argument parser created by the user that accepts additional arguments
                                  from the command line, outside the default arguments.
    :return: args: (argparse.Arguments) Command line arugments
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("--model", type=str, default='', required=True,
                        help='Name of the model to be used. A model dictionary must be provided to DLTrainer '
                             'for looking up the model class.')
    parser.add_argument("--data_dir", type=str, default="data", help="The path to the data folder")
    parser.add_argument("--save_dir", type=str, default="save", help="The path where the model and log will be saved.")
    parser.add_argument("--run_name", type=str, default='baseline', help="Name of your model training run.")
    parser.add_argument('--recompute-features', action='store_true',
                        help="Whether to recompute dataset features if they exist. "
                             "This argument can be used by your Dataset class.")
    parser.add_argument('--load_pretrained', action='store_true', help='Whether to load pretrained model')
    parser.add_argument('--pretrained_checkpoint', type=str, default='',
                        help="Directory of pretrained model. Required if load_pretrained is included.")
    parser.add_argument('--do_train', action='store_true', help="Whether to run training.")
    parser.add_argument('--do_eval', action='store_true', help="Whether to run evaluation.")
    parser.add_argument('--do_test', action='store_true', help="Whether to run test.")
    parser.add_argument('--no_eval_during_training', action='store_true',
                        help='Whether to block evaluation during training.')

    parser.add_argument('--load_optimizer', action='store_true', help='Load saved optimizer from pretrained-checkpoint')
    parser.add_argument('--load_scheduler', action='store_true', help='Load saved scheduler from pretrained-checkpoint')

    parser.add_argument('--train_batch_size', type=int, default=16,
                        help='Batch size for training, and evaluation if eval_batch_size=0')
    parser.add_argument('--eval_batch_size', type=int, default=0, help='Batch size for evaluation')
    parser.add_argument('--num_train_epochs', type=int, default=1)
    parser.add_argument('--gradient_accumulation_steps', type=int, default=1,
                        help='Number of updates steps to accumulate before performing backward & update steps.')
    parser.add_argument('--lr', type=float, default=5e-5, help='Initial learning rate for Adam.')
    parser.add_argument('--weight_decay', type=float, default=0.0, help='Weight decay if we apply some')
    parser.add_argument('--adam_epsilon', type=float, default=1e-8, help='Epsilon for Adam optimizer')
    parser.add_argument('--max_grad_norm', type=float, default=1.0, help='Max gradient norm')
    parser.add_argument('--warmup_steps', type=int, default=0, help='Linear warmup over warmup_steps')

    parser.add_argument('--seed', type=int, default=42)
    parser.add_argument('--eval_every', type=int, default=5000)
    parser.add_argument('--logging_steps', type=int, default=1000, help='Log every X update steps')

    parser.add_argument('--no_early_stopping', action='store_true',
                        help='Prevent trainer from stopping early when model performance converges on Dev set.')
    parser.add_argument('--early_stopping_steps', type=int, default=10,
                        help='Stop training early if model does not exceed --early_stopping_tol for X steps.')
    parser.add_argument('--early_stopping_tol', type=float, default=1e-5,
                        help='Stop training early if model does not exceed X for --early_stopping_steps steps.')

    parser.add_argument('--no_cuda', action='store_true', help='Avoid using CUDA when available')

    parser.add_argument('--fp16', action='store_true',
                        help='Whether to use 16-bit (mixed) precision (through NVIDIA apex) instead of 32-bit')
    parser.add_argument('--fp16_opt_level', type=str, default='01',
                        help='For fp16: Apex AMP optimization level selected in options. '
                             'See details at https://nvidia.github.io/apex/amp.html')
    parser.add_argument('--local_rank', type=int, default=-1, help="For distributed training: local_rank")
    parser.add_argument('--node_index', type=int, default=-1, help='node index if multi-node_running')
    parser.add_argument('--gpu_per_node', type=int, default=-1, help='num of gpus per node')

    if additional_arg_parser is not None:
        parser = additional_arg_parser(parser)

    args = parser.parse_args()
    return args


def train_setup(additional_arg_parser=None):
    """Parsing the training arguments from the command line.
    Setups up GPU training if applicable.
    Creates a logger for training/evaluation.

    :param additional_arg_parser: (def) A custom argument parser created by the user that accepts additional arguments
                                  from the command line, outside the default arguments.
    :return: args: (argparse.Arguments) Command line arugments
    :return: logger: (logging.Logger) Logger instance for logging events.

    """
    args = parse_input_arguments(additional_arg_parser)
    if args.do_eval or args.do_test:
        args.load_pretrained = True
    if args.load_pretrained and args.pretrained_checkpoint == '':
        raise ValueError('Must provide --pretrained_checkpoint when using --load_pretrained')
    if args.eval_batch_size == 0:
        args.eval_batch_size = args.train_batch_size
    if args.load_pretrained:
        args.save_dir = "/".join(args.pretrained_checkpoint.split('/')[:-1])
    else:
        args.save_dir = get_save_dir(args.save_dir, args.run_name)
        if not os.path.exists(args.save_dir):
            os.makedirs(args.save_dir)
    args.start_epoch = 0
    args.start_step = 0

    split_name = 'train' if args.do_train else 'validation' if args.do_eval else 'test'
    logger = get_logger(args.save_dir, 'log_train')

    logger.info("local_rank: %d, node_index: %d, gpu_per_node: %d"%(args.local_rank, args.node_index, args.gpu_per_node))
    # Setup CUDA, GPU & distributed training
    if args.local_rank == -1 or args.no_cuda:
        device = torch.device("cuda" if torch.cuda.is_available() and not args.no_cuda else "cpu")
        args.n_gpu = torch.cuda.device_count()
    else:  # Initializes the distributed backend which will take care of sychronizing nodes/GPUs
        torch.cuda.set_device(args.local_rank)
        device = torch.device("cuda", args.local_rank)
        torch.distributed.init_process_group(backend='nccl')
        args.local_rank += args.node_index * args.gpu_per_node
        args.n_gpu = 1
    args.device = device

    logger.info("Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s, world size: %s",
                   args.local_rank, device, args.n_gpu, bool(args.local_rank != -1), args.fp16,
                   torch.distributed.get_world_size() if args.local_rank != -1 else 1)

    set_seed(args)

    return args, logger


def get_logger(log_dir, name):
    """Get a `logging.Logger` instance that prints to the console
    and an auxiliary file.

    :param log_dir: (str) Directory in which to create the log file.
    :param name: (str)Name to identify the logs.
    :return: logger: (logging.Logger) Logger instance for logging events.
    """
    class StreamHandlerWithTQDM(logging.Handler):
        """Let `logging` print without breaking `tqdm` progress bars.
        See Also:
            > https://stackoverflow.com/questions/38543506
        """
        def emit(self, record):
            try:
                msg = self.format(record)
                tqdm.write(msg)
                self.flush()
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                self.handleError(record)

    # Create logger
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # Log everything (i.e., DEBUG level and above) to a file
    log_path = os.path.join(log_dir, f'{name}.txt')
    file_handler = logging.FileHandler(log_path)
    file_handler.setLevel(logging.DEBUG)

    # Log everything except DEBUG level (i.e., INFO level and above) to console
    console_handler = StreamHandlerWithTQDM()
    console_handler.setLevel(logging.INFO)

    # Create format for the logs
    file_formatter = logging.Formatter('[%(asctime)s] %(message)s',
                                       datefmt='%m.%d.%y %H:%M:%S')
    file_handler.setFormatter(file_formatter)
    console_formatter = logging.Formatter('[%(asctime)s] %(message)s',
                                          datefmt='%m.%d.%y %H:%M:%S')
    console_handler.setFormatter(console_formatter)

    # add the handlers to the logger
    logger.addHandler(file_handler)
    logger.addHandler(console_handler)

    return logger


def get_save_dir(base_dir, name, id_max=100):
    """Creates a directory name for a given model name. Adds a numeric suffix to distinguish models
    under the same name.

    :param base_dir: (str) base save directory
    :param name: (str) name of the model
    :param id_max: (int) maximum number of models under the same name
    :return:
    """
    for uid in range(1, id_max):
        save_dir = os.path.join(base_dir, f'{name}-{uid:02d}')
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
            return save_dir

    raise RuntimeError('Too many save directories created with the same name. \
                       Delete old save directories or use another name.')


from . import trainer_utils

import dill
import os
import torch
from torch.optim import AdamW
from torch.utils.data import DataLoader, SequentialSampler, RandomSampler
from torch.utils.data.distributed import DistributedSampler
from transformers import get_linear_schedule_with_warmup
from tensorboardX import SummaryWriter
from tqdm import tqdm


class DLTrainer:
    """
    Base trainer class for training deep learning models.
    """
    def __init__(self, MODELS, metric_fn=None, additional_arg_parser=None):
        """Constructor

        :param MODELS:
        :param additional_arg_parser:
        :param metric_fn
        """
        self.args, self.logger = trainer_utils.train_setup(additional_arg_parser)
        self.metric_fn = metric_fn

        # Barrier to make sure only the first process in distributed training downloads model & vocab
        if self.args.local_rank not in [-1, 0]:
            torch.distributed.barrier()

        self.model_checkpoint = os.path.join(self.args.save_dir, 'checkpoint') if self.args.pretrained_checkpoint == '' \
            else self.args.pretrained_checkpoint

        classes = MODELS[self.args.model]
        if len(classes) == 4:
            config_class, model_class, dataset_class, tokenizer_class = classes
            # tokenizer must be pretrained and placed in model_checkpoint folder
            tokenizer = tokenizer_class.from_pretrained(self.model_checkpoint)
        elif len(classes) == 3:
            config_class, model_class, dataset_class = classes
        else:
            raise ValueError(f"MODELS class list must contain either 4 elements for NLP or 3 elements otherwise. "
                             "Received {len(classes)} elements.")
        if self.args.load_pretrained:
            if callable(getattr(model_class, "from_pretrained", None)):
                self.config = config_class.from_pretrained(self.model_checkpoint)
                self.model = model_class.from_pretrained(self.model_checkpoint)
            else:
                self.model, self.config = self.load_model(model_class)
        else:
            self.config = config_class(self.args)
            self.model = model_class(self.config)

        num_params = sum([p.numel() for p in self.model.parameters()])
        self.logger.info(f"Model has a total of {num_params} trainable parameters.")

        # End of distributed training barrier
        if self.args.local_rank == 0:
            torch.distributed.barrier()

        self.logger.info("Training/evaluation parameters %s", self.args)

        # Training
        if self.args.do_train:
            if not os.path.exists(self.args.save_dir):
                os.makedirs(self.args.save_dir)

            train_dataset = dataset_class(self.args, 'train')
            train_sampler = RandomSampler(train_dataset)
            train_dataloader = DataLoader(train_dataset, sampler=train_sampler, batch_size=self.args.train_batch_size)
            val_dataloader = None
            if not self.args.no_eval_during_training:
                pass
                val_dataset = dataset_class(self.args, 'dev')
                val_sampler = SequentialSampler(val_dataset) if self.args.local_rank == -1 else DistributedSampler(
                    val_dataset)
                val_dataloader = DataLoader(val_dataset, sampler=val_sampler, batch_size=self.args.eval_batch_size)
            self.train(train_dataloader, val_dataloader)

        if self.args.do_eval or self.args.do_test:
            eval_dataset = dataset_class(self.args, 'dev' if self.args.do_eval else 'test')
            eval_sampler = SequentialSampler(eval_dataset) if self.args.local_rank == -1 else DistributedSampler(
                eval_dataset)
            eval_dataloader = DataLoader(eval_dataset, sampler=eval_sampler, batch_size=self.args.eval_batch_size)
            results = self.evaluate(eval_dataloader)
            print(results)

    def _default_optimizer(self, t_total):
        no_decay = ['bias', 'LayerNorm.weight']
        optimizer_grouped_parameters = [
            {'params': [p for n, p in self.model.named_parameters() if not any(nd in n for nd in no_decay)],
             'weight_decay': self.args.weight_decay},
            {'params': [p for n, p in self.model.named_parameters() if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
        ]
        optimizer = AdamW(optimizer_grouped_parameters, lr=self.args.lr, eps=self.args.adam_epsilon)
        scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps=self.args.warmup_steps,
                                                    num_training_steps=t_total)
        return optimizer, scheduler

    def create_optimizer(self, t_total):
        """Create optimizer for training.

        Should be overwritten for special cases.

        :param model: (Any) model to train
        :param t_total: (int) total number of training steps
        :return:
        """
        self.optimizer, self.scheduler = self._default_optimizer(t_total)

    def _default_process_batch(self, batch):
        inputs = [x.to(self.args.device) for x in batch[0] if x is not None]
        labels = batch[1].to(self.args.device)
        return inputs, labels

    def process_batch(self, batch):
        """Process data batch for model input
        Sends each tensor to the model device.

        Should be overwritten for special cases.

        :param batch: (list) The batch from the dataloader
        :return: batch: (list) The batch to be input into the model, with each tensor on the model device
        """
        return self._default_process_batch(batch)

    def train_step(self, inputs, labels):
        """Perform a training step.

        Can be overwritten for special cases.

        :param inputs: (list) Batch inputs
        :return: loss: (Any) training loss
        """
        inputs += [labels]
        outputs = self.model(*inputs)
        loss = outputs[0]

        # mean() to average on multi-gpu parallel training
        if self.args.n_gpu > 1:
            loss = loss.mean()
        if self.args.gradient_accumulation_steps > 1:
            loss = loss / self.args.gradient_accumulation_steps

        self.optimizer.zero_grad()
        if self.args.fp16:
            with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                scaled_loss.backward()
            torch.nn.utils.clip_grad_norm_(amp.master_params(self.optimizer), self.args.max_grad_norm)
        else:
            loss.backward()
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.args.max_grad_norm)

        return loss

    def train(self, train_dataloader, val_dataloader):
        self.model.to(self.args.device)
        if self.args.local_rank in [-1, 0]:
            tbx_path = self.args.save_dir + '/tensorboard'
            if not os.path.exists(tbx_path):
                os.makedirs(tbx_path)
            tbx = SummaryWriter(tbx_path)

        if self.args.local_rank not in [-1, 0]:
            torch.distributed.barrier()

        # Prepare optimizer and schedule (linear warmup and decay)
        total_examples = len(train_dataloader.dataset) * (
            torch.distributed.get_world_size() if self.args.local_rank != -1 else 1)
        batch_size = self.args.train_batch_size * self.args.gradient_accumulation_steps * (
            torch.distributed.get_world_size() if self.args.local_rank != -1 else 1)
        t_total = total_examples // batch_size * self.args.num_train_epochs

        self.create_optimizer(t_total)
        optimizer_checkpoint = os.path.join(self.model_checkpoint, 'optimizer.pt')
        scheduler_checkpoint = os.path.join(self.model_checkpoint, 'scheduler.pt')

        if self.args.load_optimizer and os.path.exists(optimizer_checkpoint):
            self.logger.info(f"Loading optimizer from {optimizer_checkpoint}")
            self.optimizer.load_state_dict(torch.load(optimizer_checkpoint, map_location="cpu"))
        if self.args.load_scheduler and os.path.exists(scheduler_checkpoint):
            self.logger.info(f"Loading scheduler from {scheduler_checkpoint}")
            self.scheduler.load_state_dict(torch.load(scheduler_checkpoint, map_location="cpu"))

        if self.args.local_rank == 0:
            torch.distributed.barrier()

        if self.args.fp16:
            try:
                from apex import amp
            except ImportError:
                raise ImportError("Please install apex from https://www.github.com/nvidia/apex to use fp16 training.")
            self.model, self.optimizer = amp.initialize(self.model, self.optimizer, opt_level=self.args.fp16_opt_level)

        # multi-gpu training (should be after apex fp16 initialization)
        if self.args.n_gpu > 1:
            self.model = torch.nn.DataParallel(self.model)

        # Distributed training (should be after apex fp16 initialization)
        if self.args.local_rank != -1:
            model = torch.nn.parallel.DistributedDataParallel(self.model,
                                                              device_ids=[self.args.local_rank % self.args.gpu_per_node],
                                                              output_device=self.args.local_rank % self.args.gpu_per_node,
                                                              find_unused_parameters=True)

        # Train!
        self.logger.info("***** Running training *****")
        self.logger.info("  Num examples = %d", total_examples)
        self.logger.info("  Num epoch = %d", t_total * batch_size // total_examples)
        self.logger.info("  Instantaneous batch size per GPU = %d", self.args.train_batch_size)
        self.logger.info("  Total train batch size (w. parallel, distributed & accumulation) = %d", batch_size)
        self.logger.info("  Gradient Accumulation steps = %d", self.args.gradient_accumulation_steps)
        self.logger.info("  Total optimization steps = %d", t_total)

        global_step = self.args.start_step
        tr_loss, logging_loss, avg_loss, tr_nb = 0.0, 0.0, 0.0, global_step
        eval_steps_no_increase = 0
        best_score = -float('inf')
        for epoch in range(self.args.start_epoch, self.args.num_train_epochs):
            for step, batch in enumerate(train_dataloader):
                # batch should be tuple of length 2 (inputs, gold_labels)
                self.model.train()
                self.optimizer.zero_grad()
                inputs, labels = self.process_batch(batch)
                loss = self.train_step(inputs, labels)
                tr_loss += loss.item()

                if (step +1) % self.args.gradient_accumulation_steps == 0:
                    self.optimizer.step()
                    self.scheduler.step()
                    global_step += 1

                    avg_loss = round((tr_loss - logging_loss) / (global_step - tr_nb), 4)
                    if global_step % self.args.logging_steps == 0:
                        self.logger.info(" steps: %s loss; %s", global_step, avg_loss)
                    if self.args.local_rank in [-1, 0] and self.args.logging_steps > 0 \
                        and global_step % self.args.logging_steps == 0:
                        # log metrics
                        tbx.add_scalar('lr', self.scheduler.get_last_lr()[0], global_step)
                        tbx.add_scalar('loss', (tr_loss - logging_loss) / self.args.logging_steps)
                        logging_loss = tr_loss
                        tr_nb = global_step
                    if self.args.local_rank in [-1, 0] and self.args.eval_every > 0 \
                            and global_step % self.args.eval_every == 0:
                        # save model checkpoint
                        if not self.args.no_eval_during_training:
                            results = self.evaluate(val_dataloader)
                            for key, value in results.items():
                                tbx.add_scalar('eval_{}'.format(key), value, global_step)
                                self.logger.info(" %s = %s, best_score = %s", key, round(value, 4), round(best_score, 4))

                            score = results[list(results.keys())[0]]

                            if score < best_score + self.args.early_stopping_tol:
                                eval_steps_no_increase += 1
                            else:
                                eval_steps_no_increase = 0

                            if score > best_score:
                                best_score = score
                                self.save_model()
                                self.logger.info("Saving model checkpoint to %s", self.model_checkpoint)

                            if eval_steps_no_increase > self.args.early_stopping_steps:
                                self.logger.info(f"Stopping at step {global_step} in epoch {epoch}. "
                                                 f"Validation did not improve by {self.args.early_stopping_tol} for "
                                                 f"more than {self.args.early_stopping_steps} steps.")
                                return
                        else:
                            self.save_model()
                            self.logger.info("Saving model checkpoint to %s", self.model_checkpoint)

    def eval_step(self, inputs, labels=None):
        """Perform a eval step.

        Can be overwritten for special cases.

        :param inputs: (list) Batch inputs
        :param labels: (torch.tensor) Gold labels for task, if provided
        :return: loss: (Any) training loss
        """
        if labels is not None:
            inputs += [labels]
        outputs = self.model(*inputs)
        return outputs

    def evaluate(self, eval_dataloader):
        # Eval!
        split = 'validation' if self.args.do_train or self.args.do_eval else 'test'
        self.logger.info("***** Running evaluation {} *****".format(split))
        self.logger.info("  Num examples = %d", len(eval_dataloader.dataset))
        self.logger.info("  Batch size = %d", self.args.eval_batch_size)
        eval_loss = 0.0
        nb_eval_steps = 0
        self.model.eval()

        all_outputs = []
        all_labels = []
        for batch in tqdm(eval_dataloader):
            # batch should be tuple of length 2 (inputs, gold_labels)
            inputs, labels = self.process_batch(batch)
            with torch.no_grad():
                outputs = self.eval_step(inputs, labels)
            loss = outputs[0]
            if loss is not None:
                eval_loss += loss.mean().item()
            nb_eval_steps += 1

            if labels is not None:
                all_outputs.append(outputs[1].cpu())
                all_labels.append(labels.cpu())

        all_outputs = torch.cat(all_outputs, dim=0)
        all_labels = torch.cat(all_labels, dim=0)

        eval_loss = eval_loss / nb_eval_steps
        if self.metric_fn is None:
            results = {'eval_loss': -eval_loss}
        else:
            results = self.metric_fn(all_outputs, all_labels)
        return results

    def load_model(self, model_class):
        config = dill.load(open(f'{self.model_checkpoint}/config.pt', 'rb'))
        model = model_class(config)
        model.load_state_dict(torch.load(f'{self.model_checkpoint}/pytorch_model.bin'))
        return model, config

    def save_model(self):
        """Save a checkpoint in the output directory"""
        if os.path.isfile(self.model_checkpoint):
            return
        os.makedirs(self.model_checkpoint, exist_ok=True)

        # Only save the model itself if we are using distributed training
        model_to_save = self.model.module if hasattr(self.model, "module") else self.model

        state_dict = model_to_save.state_dict()
        torch.save(state_dict, f'{self.model_checkpoint}/pytorch_model.bin')

        # save config
        dill.dump(self.config, open(f'{self.model_checkpoint}/config.pt', 'wb'))
        # save optimizer
        torch.save(self.optimizer.state_dict(), f'{self.model_checkpoint}/optimizer.pt')
        # save scheduler
        torch.save(self.scheduler.state_dict(), f'{self.model_checkpoint}/scheduler.pt')

